"""
Karma - handle karma (++ and --) tracking
Copyright (C) 2010  Brian S. Stephan

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import sqlite3

from Module import Module

__author__ = "Mike Bloy <mike@bloy.org>"
__date__ = "$Oct 23, 2010 11:12:33 AM$"

class Karma(Module):

    def __init__(self, config, server, modlist):
        """
        Upon creation, determine the save file location
        """

        Module.__init__(self, config, server, modlist)

        pattern = "(?:(\S+)|\((.+)\))"
        karmapattern = pattern + '(\+\+|--|\+-|-\+)' + '(\s+|$)'
        querypattern = '^rank\s+(.*)'

        self.karmare = re.compile(karmapattern)
        self.queryre = re.compile(querypattern)

    def db_init(self):
        # need to init the database if karma tables don't already exist
        version = self.db_module_registered(self.__class__.__name__)
        if (version == None):
            # have to create the database tables
            conn = self.get_db()
            try:
                conn.execute('''
                   CREATE TABLE karma_log (
                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                        key TEXT NOT NULL,
                        delta INTEGER NOT NULL,
                        who TEXT NOT NULL,
                        userhost TEXT NOT NULL,
                        karmatime TEXT DEFAULT CURRENT_TIMESTAMP
                    )''')
                conn.execute('CREATE INDEX karma_log_key_ix ON karma_log (key)')
                conn.execute('CREATE INDEX karma_log_who_ix ON karma_log (who)')
                conn.execute('''
                    CREATE VIEW karma_values AS
                    SELECT key, SUM(delta) AS value
                    FROM karma_log
                    GROUP BY key''')

                sql = 'INSERT INTO drbotzo_modules VALUES (?,?)'
                conn.execute(sql, (self.__class__.__name__, 1))
                conn.commit()
            except sqlite3.Error as e:
                conn.rollback()
                print("sqlite error: " + str(e))
                raise

    def do(self, connection, event, nick, userhost, replypath, what, admin_unlocked):
        """look for karma strings at the start of messages"""

        if (self.karmare.search(what)):
            return self.handle_karma_change(connection, nick, userhost, replypath, what)
        elif (self.queryre.search(what)):
            return self.handle_karma_query(connection, nick, userhost, replypath, what)

    def handle_karma_change(self, connection, nick, userhost, replypath, what):
        """
        handle the karma change and storage.
        """
        match = self.karmare.search(what)
        key = match.group(1) if match.group(1) else match.group(2)
        value = match.group(3)
        if (value == '++'):
            return self.karma_modify(key, 1, nick, userhost)
        elif (value == '--'):
            return self.karma_modify(key, -1, nick, userhost)
        elif (value == '+-'):
            self.karma_modify(key, 1, nick, userhost)
            return self.karma_modify(key, -1, nick, userhost)
        elif (value == '-+'):
            self.karma_modify(key, -1, nick, userhost)
            return self.karma_modify(key, 1, nick, userhost)

    def karma_modify(self, key, value, nick, userhost):
        """
        Go out to the database and update the karma value.
        """

        conn = self.get_db()
        try:
            sql = '''
            INSERT INTO karma_log (key, delta, who, userhost)
            VALUES (?, ?, ?, ?)
            '''
            conn.execute(sql, (key, value, nick, userhost))
            conn.commit()
        except sqlite3.Error as e:
            conn.rollback()
            return self.reply(connection, replypath, "sqlite error: " + str(e))

    def handle_karma_query(self, connection, nick, userhost, replypath, what):
        match = self.queryre.search(what)
        key = match.group(1)

        conn = self.get_db()
        reply = '{nick}: {key} has no karma'.format(nick=nick, key=key)
        try:
            query = '''
            SELECT value
            FROM karma_values
            WHERE key = :key
            '''
            value = conn.execute(query, {'key': key}).fetchone()

            if (value != None):
                query = '''
                SELECT count(*) FROM karma_values WHERE value > :value
                '''
                rank = conn.execute(query, {'value': value[0]}).fetchone()
                rank = rank[0] + 1;

                reply = '{nick}: {key} has {value[0]!s} points of karma (rank {rank})'.format(
                    nick=nick, key=key, value=value, rank=rank)
        except sqlite3.Error as e:
            return self.reply(connection, replypath, "sqlite error: " + str(e))

        return self.reply(connection, replypath, reply)

if __name__ == "__main__":
    print "Hello World"

# vi:tabstop=4:expandtab:autoindent
